/*
 * RANDOM.C
 * Random routines for FRP games
 * By Joshua Levy 
 *
%   cc -g -c random.c
%t  cc -g -DTEST -o rndtest random.c -lm
 *
 * This modual has five "functional" entry points:
 *      int ra_roll(str) char *str ;
 *      char *ra_card() 
 *      int ra_random(n) int n ;
 *      int ra_float()
 *      int ra_query(str,quest) char *str ; int quest ;
 *      int ra_gasdev() ;
 *      int ra_range(low,middle,high) int low, middle, high ;
 *
 * and two "state" entry points:
 *
 *      void ra_seed(seed) long seed ;
 *          This initializes the random number generator, if it is not called
 *          the random number generator will be in an unknown state.  Passing
 *          zero will usually generate a repeatable series of numbers and is 
 *          good for debugging.  A different number should be used "for real".
 *      int ra_logging(state,filename) int state ; char *filename ;
 *          This allows the program to log all of the random numbers it
 *          generates, and also use random number from the log, if desired.
 *
 * Configuation
 *      I have used random and srandom, since they are available on my
 *      UNIX system, and are better than rand and srand which are more
 *      widely available. 
 */

#include <sys/types.h>
#include <stdio.h>
#include <assert.h>
#include <values.h>      /* only used for MAXINT */

#include "random.h"

static int ra_state = RA_NORMAL ;
static FILE *ra_log = NULL , *ra_play = NULL ;

/*
 * If you want to use a different random number generator, you
 * should only need to change things on this page, from here down.
 */

extern long random() ;

int ra_random(n) unsigned int n ;
{
    int result ;
    int s ;

    if (ra_state & RA_PLAY) {
	assert(ra_play) ;
	s = fscanf(ra_play,"%d ",&result) ;
	if (s==EOF) {
#ifndef REPEAT
	    printf("ra_random: prematurely reached end of logging file.\n") ;
	    printf("           continuing with random numbers\n") ;
	    fclose(ra_play) ;
	    ra_play = (FILE*)NULL ;
	    ra_state ^= RA_PLAY ;
	    result = random()%n+1 ;
#else
	    fseek(ra_play,0L,0) ;
	    s = fscanf(ra_play,"%d ",&result) ;
#endif
	}
    } else {
	result = random()%n+1 ;
    }
    if (ra_state & RA_LOG) {
	assert(ra_log) ;
	fprintf(ra_log,"%d ",result) ;
    }    
    return result ;
}

void ra_seed(seed) int seed ;
{
    srandom(seed) ;
}

int ra_logging(state,filename) int state ; char *filename ;
{
    char buf[100] ;
    time_t made ;

    if ( state!=RA_NORMAL && state!=RA_PLAY && state!=RA_LOG )
	return RA_E_STATE ;
    if (state>0)
	ra_state |= state ;
    else if (state<0)
	ra_state ^= state ;
    else /*state==0*/
	ra_state = state ;
    switch (state) {  /* remember the switch fall through rules */
      case RA_NORMAL:
	if (ra_play) { fclose(ra_play) ; ra_play = (FILE*)NULL ; }
	if (ra_log) { fclose(ra_log) ; ra_log = (FILE*)NULL ; }
	break ;
      case -RA_PLAY:
	if (ra_play) { fclose(ra_play) ; ra_play = (FILE*)NULL ; }
	break ;
      case -RA_LOG:
	if (ra_log) { fclose(ra_log) ; ra_log = (FILE*)NULL ; }
	break ;
      case RA_PLAY:
	ra_play = fopen(filename,"r") ;
	if (!ra_play)
	    return RA_E_NOFILE ;
	fgets(buf,sizeof(buf),ra_play) ;
	/* right now, no check is made */
	break ;
      case RA_LOG:
	ra_log = fopen(filename,"w") ;
	if (!ra_log)
	    return RA_E_NOFILE ;
	made = time((time_t*)0) ;
	fprintf(ra_log,"ra_log_file %d %s",RA_VERSION,ctime(&made)) ;
	break ;
    }
    return ra_state ;
}

char *ra_card() 
{
    static char ret[10] ;
    static char allsuites[5] = "cdhs" ;

    int card = ra_random(12) ;
    int suit = ra_random(4)-1 ;

    sprintf(ret,"%d%c",card,allsuites[suit]) ;
    return ret ;
}

/*
 * Bug: first number is required: 1d20 , 1d100, etc.  
 * (d20 will fail for example)
 */
ra_roll(t) char *t ;
{
    int num=0, del=0 ;
    unsigned int type=0 ;
    char dice='d', sign='+' ;
    int v=0 ;
    int i ;
    int s ;

    s = sscanf(t,"%d%c%u%c%d",&num,&dice,&type,&sign,&del) ;
    if (s==3) {
	sign = '+' ;
	del = 0 ;
    } 
    for(i=0 ; i<num ; i++) 
	v += ra_random(type) ;
    if (sign=='+')
	v += del ;
    else
	v -= del ;
    return v ;
}

int ra_query(roll,what) char *roll ; int what ;
{
    int num=1, del=0 ;
    unsigned int type=0 ;
    char dice='d', sign='+' ;
    int v=0 ;
    int i ;
    int s ;
    int legal=1 ;

    if (roll[0]=='d' || roll[0]=='D') {
	s = sscanf(roll,"%c%u%c%d",&dice,&type,&sign,&del) ;
	s += 1 ;
    } else {
	s = sscanf(roll,"%d%c%u%c%d",&num,&dice,&type,&sign,&del) ;
    }
    if (s!=5 && s!=3) {
	legal = 0 ;
    }
    if (s==3) {
	sign = '+' ;
	del = 0 ;
    } else {
	if (sign=='-')
	    del = -del ;
    }
    if (dice!='d' && dice!='D') {
	legal = 0 ;
    }
    switch (what) {
      case RA_MAX:
	return num*type+del ;
      case RA_MIN:
	return num+del ;
      case RA_AVE:
	return (num*(type/2))+del ;
      case RA_LEGAL:
	return legal ;
      default:
	return RA_E_BADQUERY ;
    }
    /*NOTREACHED*/
}

float ra_float() 
{
    float x ;
    x = (ra_random(MAXLONG)-1)/(MAXLONG+1.0) ;
    return x ;
}

/*
 * The following routine is taken from SOFTWARE RECIPIES IN C
 * By Press, Flannery, and a cast of thousands.
 * Look on page 217 (section 7.2)
 */
float ra_gasdev()
{
    static int iset=0 ;
    static float gset ;
    float fac,r,v1,v2 ;

    if (iset==0) {
	do {
	    v1 = 2.0*ra_float()-1.0 ;
	    v2 = 2.0*ra_float()-1.0 ;
	} while (r >= 1.0) ;
	fac = sqrt(-2.0*log(r)/r) ;
	gset = v1*fac ;
	iset = 1 ;
	return v2*fac ;
    } else {
	iset=0 ;
	return gset ;
    }
}

ra_range(low,middle,high) int low ; int middle ; int high ;
{
    float temp ;

    temp = ra_gasdev() ;
    temp = temp/3.0 ;
    if (temp==0.0) return middle ;
    if (temp>0.0) return (int) (middle+(temp*(high-middle))) ;
    if (temp<0.0) return (int) (middle-(temp*(middle-low))) ;
    return middle ;
}
#if 0
/*
 * It should be used when you need to implement rules like: "the player
 * can allocate 4 skill points amoung the following six rules."  This
 * would be implemented with a call like ra_allocate(4,6,"flat").
 *     total is the total number of things to be allocated.
 *     bins is the number of places things can go.
 *     method is the algorithm you want to use to do the allocation.
 *         legal methods are: any dice roll, at floating point text. 
 * This routine returns an array of integers of bins size, which is static
 * and will be reused with the next call.
 */
int *ra_allocate(total,bins,method) int total ; int bins ; char *method ;
{
    ;
}
#endif

#if TEST

#define CHECK_ARG_COUNT(num)       \
    if (num!=count-1) {            \
        printf("rndtest (%s): got %d args, but expeced %d args\n", \
	       comm,count-1,num) ; \
        continue ;     }           \

main()
{
    char buf[1000] ;
    char arg1[20], arg2[20], arg3[20], arg4[20] ;
    char comm[20] ;
    int count ;
    int echo = 0 ;
    int i,m ;

    while (fgets(buf,sizeof(buf),stdin)) { 
	if (echo)
	    printf("%s",buf) ;
	count = sscanf(buf,"%s %s %s %s %s",comm,arg1,arg2,arg3,arg4) ;
	if (!strcmp(comm,"quit")) {
	    exit(0) ;
	} else if (!strncmp(comm,"#",1)) {
	    continue ;
	} else if (!strcmp(comm,"echo")) {
	    echo = 1 ;
	} else if (!strcmp(comm,"roll")) {
	    CHECK_ARG_COUNT(2) ;
	    m = atoi(arg1) ;
	    for(i=0 ; i<m ; i++)
		printf("%d ",ra_roll(arg2)) ;
	    printf("\n") ;
	} else if (!strcmp(comm,"query")) {
	    CHECK_ARG_COUNT(2) ;
	    m = atoi(arg1) ;
	    if (!m) {
		if (!strcmp(arg1,"RA_MIN"))
		    m = RA_MIN ;
		else if (!strcmp(arg1,"RA_MAX"))
		    m = RA_MAX ;
		else if (!strcmp(arg1,"RA_AVE"))
		    m = RA_AVE ;
		else if (!strcmp(arg1,"RA_LEGAL"))
		    m = RA_LEGAL ;
	    }
	    printf("%d\n",ra_query(arg2,m)) ;
	} else if (!strcmp(comm,"card")) {
	    CHECK_ARG_COUNT(1) ;
	    m = atoi(arg1) ;
	    for(i=0 ; i<m ; i++)
		printf("%s ",ra_card()) ;
	    printf("\n") ;
	} else if (!strcmp(comm,"seed")) {
	    CHECK_ARG_COUNT(1) ;
	    ra_seed(atoi(arg1)) ;
	} else if (!strcmp(comm,"log")) {
	    CHECK_ARG_COUNT(2) ;
	    ra_logging(atoi(arg1),arg2) ;
	} else if (!strcmp(comm,"float")) {
	    CHECK_ARG_COUNT(1) ;
	    m = atoi(arg1) ;
	    for(i=0 ; i<m ; i++) {
		printf("%g ",ra_float()) ;
		if (!(i%5)) printf("\n") ;
            }
	    printf("\n") ;
	} else if (!strcmp(comm,"gasdev")) {
	    CHECK_ARG_COUNT(1) ;
	    m = atoi(arg1) ;
	    for(i=0 ; i<m ; i++)
		printf("%g ",ra_gasdev()) ;
	    printf("\n") ;
	} else if (!strcmp(comm,"range")) {
	    CHECK_ARG_COUNT(4) ;
	    m = atoi(arg1) ;
	    for(i=0 ; i<m ; i++)
		printf("%g ",ra_range(atoi(arg2),atoi(arg3),atoi(arg4))) ;
	    printf("\n") ;
	} else {
	    printf("Illegal test command.  Legal commands are:\n") ;
	    printf("quit, roll, card, seed, float, log, query, gasdev, and range.\n") ;
	    printf("Note that these are all lower case names.\n") ;
	}
    }
    printf("\n") ;
    exit(0) ;
}

#endif

