/*
 * TABLE.C
 * Code supporting rolling on tables
 * By Joshua levy  
 *
 * COMPILING
 *
 * This line creates a simple test program:
%  cc -g -DTEST -o table table.c tablesimple.o tablemulti.o random.o
 *
 * This line creates an 'o' file to include in other programs:
%o cc -g -c table.c 
 *
 * INTERFACE
 *   There is one data type:
 *     table_t
 *   and seven interfaces in this file:
 *     tb_read(table,file) table_t **table ; char *file ;
 *     tb_use(table,...,result) table_t *table ; char **result ;
 *     tb_free(table) table_t *table ;
 *     tb_print(table,way) table_t *table ; int way ;
 *     tb_many(table,count,...,result) table_t *table ; count int ; char **result ;
 *     tb_choose(table,which,...,result) table_t *table ; int which ; char **result ;
 *     tb_back(table,lookfor,result) table_t *table ; char *lookfor ; char **result ;
 *
 * ADDING NEW TABLES
 *    This is supposed to be the framework for an object oriented package,
 *    which support many different types of tables.  This is the (untested) 
 *    procedure for adding a new table type (refered to as a foo table):
 *        create tablefoo.h
 *            Containing declarations for the functions in tablefoo.c.
 *        create tablefoo.c
 *            It should contain at least three entry points: tb_foo_read,
 *            tb_foo_free, and tb_foo_use.  I also suggest tb_foo_print,mostly
 *            for debugging.
 *        modify table.h
 *            So that it includes tablefoo.h.
 *        modify table.c
 *            So that the tba_* tables contain pointers the new routines.
 *    Note: I do not recommend adding new tables at this time.  I'm hoping
 *    for a "real" release in a month or three, and I'll have better support
 *    for new table types then.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <varargs.h>

#include "random.h"
#include "table.h"

extern int illegal() ;

#define assert(ex) {if (!(ex)){fprintf(stderr,"Assertion failed: file %s, line %d\n", __FILE__, __LINE__);fprintf(stderr,"sts is %d\n",sts);exit(1);}}

extern int tb_do_query() ;

/*
 * Data Tables.
 * These arrays form a simple dispatch mechanism.
 */

/* maps type name to type number */
typedef struct { char *typ ; int t } mapper ;

mapper tba_type[] = {
    {"Simple", 1}, {"Multiple",2}, {"SimpleBig",3}, {NULL, 0},
} ;

/* table of read routines indexed by type number */
int (*tba_read[])() = {
    illegal, tb_simple_read, tb_multi_read, tb_simpbig_read,
} ;

/* table of use routines indexed by type number */
int (*tba_use[])() = {
    illegal, tb_simple_use, tb_multi_use,tb_simple_use,
} ;

/* table of print routines indexed by type number */
int (*tba_print[])() = {
    illegal, tb_simple_print, tb_multi_print,tb_simpbig_print,
} ;

/* table of choose routines indexed by type number */
int (*tba_choose[])() = {
    illegal, tb_simple_choose, tb_multi_choose,tb_simple_choose,
} ;

/* table of back routines indexed by type number */
int (*tba_back[])() = {
    illegal, tb_simple_back, tb_multi_back, tb_simple_back,
} ;

/* table of free routines indexed by type number */
int (*tba_free[])() = {
    illegal, tb_simple_free, tb_multi_free,tb_simple_free,
} ;

/* table of query routines indexed by type number */
int (*tba_query[])() = {
    illegal, tb_do_query, tb_do_query, tb_do_query,
} ;

/*
 * A couple of utility routines
 */
#ifdef NEEDDUP
char *strdup(str) char *str ;
{
    char *ret = (char*)malloc(strlen(str)+1) ;

    if (ret)
	strcpy(ret,str) ;
    return ret ;
}
#endif

int map(text,mappi) char *text ; mapper mappi[] ;
{
    int i ;
    for(i=0 ; mappi[i].typ ; i++) 
	if (!strcmp(mappi[i].typ,text))
	    return mappi[i].t ;
    return 0 ;
}

int illegal()
{
    fprintf(stderr,"Attempt to send a message to an illegal type of table.\n");
    fprintf(stderr,"Exiting program.\n") ;
    exit(-2) ;
}

int tb_do_query(tab,query,res) table_t *tab ; int query ; int *res ;
{
    return ra_query(((alltable_t*)tab)->roll,query,res) ;
}

/*
 * the read utilities for tables
 */
int tb_read(file,table) char *file ; table_t **table ; 
{
    char buf[1000] ;
    FILE *fp ;
    char *cp ;
    char *name ;
    char *type ;
    int sts=0 ;
    int t ;

    fp = fopen(file,"r") ;
    if (!fp)
	return TB_E_NOFILE ;
    
    cp = fgets(buf,sizeof(buf),fp) ;
    assert(cp) ;
    name = strtok(buf,"\t\n") ;
    type = strtok((char*)0,"\t\n") ;
    assert(type) ;
    if (!(t = atoi(type)))
	t = map(type,tba_type) ;
    assert(t) ;
    sts = tba_read[t](table,fp,name) ;
    assert(sts!=-1) ;
    (*table)->type = t ;
    FCLOSE(fp) ;
    return TB_OK ;
}

/* 
 * Dump the table in a way that can be used by tbl/troff 
 */
int tb_print(table,file,way) table_t *table ; char *file ; int way ;
{
    FILE *fp ;
    int sts=0 ;

    if (!file)
	fp = stderr ;
    else if (!file[0])
	fp = stdout ;
    else
	fp = fopen(file,"r") ;
    assert(fp) ;
    
    sts = tba_print[table->type](table,fp,way) ;
    if (file && file[0])
	FCLOSE(fp) ;
    return sts ;
}

/*
 * use a table.
 */
int tb_use(table,arg1,arg2) table_t *table ; char *arg1 ; char *arg2 ;
{
    int sts ;

    switch (table->type) {
      case 1:
      case 3:
	sts = tba_use[table->type](table,arg1) ;
	break ;
      case 2:
	sts = tba_use[table->type](table,arg1,arg2) ;
	break ;
      default:
        sts = TB_E_NOTYPE ;
    }
    return sts ;
}

/*
 * free a table.
 */
int tb_free(table) table_t *table ; 
{
    return tba_free[table->type](table) ;
}

/*
 * use a table many times.
 */
int tb_many(table,count,arg1,arg2) table_t *table ; char *arg1 ; char *arg2 ;
{
    int sts ;
    int r ;
    char *tmp ;
    char **res ;
    int got=0 ;
    int bad ;

    if (count>table->all.num)
	return TB_E_TOOMANY ;
    while (got<count) {
	/*
	 * first, we roll on the table
	 */
	switch (table->type) {
	  case 1:
	  case 3:
	    res = (char**)arg1 ;
	    sts = tb_use(table,&tmp) ;
	    break ;
	  case 2:
	    res = (char**)arg2 ;
	    sts = tb_use(table,arg1,&tmp) ;
	    break ;
	  default:
	    sts = TB_E_NOTYPE ;
	}
	if (sts!=0)
	    return sts ;
	/*
	 * second, we see if we've gotten this result before
	 */
	for(r=0,bad=0 ; r<got ; r++) {
	    if (!strcmp(res[r],tmp))
		bad=1 ;
	}
	/*
	 * finally, if not, we add it two our list
	 */
	if (!bad) {
	    res[got++] = tmp ;
	}
    }
    return sts ;
}

/*
 * choose from a table.
 */
int tb_choose(table,which,arg1,arg2) table_t *table; int which; char *arg1, *arg2 ;
{
    int sts ;

    switch (table->type) {
      case 1:
      case 3:
	sts = tba_choose[table->type](table,which,arg1) ;
	break ;
      case 2:
	sts = tba_choose[table->type](table,which,arg1,arg2) ;
	break ;
      default:
        sts = TB_E_NOTYPE ;
    }
    return sts ;
}
/*
 * use a table backwards
 */
int tb_back(table,which,va_alist) table_t *table; char *which ; va_dcl
{
    int sts ;
    char **ret ;
    char *arg ;
    va_list list ;

    va_start(list) ;
    switch (table->type) {
      case 1:
      case 3:
	ret = va_arg(list,char**) ;
	sts = tba_back[table->type](table,which,ret) ;
	break ;
      case 2:
	arg = va_arg(list,char*) ;
	ret = va_arg(list,char**) ;
	sts = tba_back[table->type](table,which,arg,ret) ;
	break ;
      default:
        sts = TB_E_NOTYPE ;
    }
    va_end(list) ;
    return sts ;
}

int tb_query(t,q,r) table_t *t ; int q ; int *r ;
{
    return tba_query[t->type](t,q,r) ;
}


#ifdef TEST
 
main(argc,argv) int argc ; char *argv[] ;
{
    table_t *table ;
    char text[100] ;
    char *result ;
    int count, i ;
    int sts=0 ;

    assert(argc==3) ;
    count = atoi(argv[2]) ;

    tb_read(argv[1],&table) ;
    assert(sts==TB_OK) ;

    sts = tb_print(table,"",0) ;
    assert(sts==TB_OK) ;

    switch (table->type) {
      case 1:
	for(i=0 ; i<count ; i++) {
	    sts = tb_use(table,&result) ;
	    assert(sts==TB_OK) ;
	    printf("%d %s\n",i+1,result) ;
	}
	break ;
      case 2:
	printf("Enter column: ") ;
	fgets(text,sizeof(text),stdin) ;
	text[strlen(text)-1] = '\0' ;
	for(i=0 ; i<count ; i++) {
	    sts = tb_use(table,text,&result) ;
	    assert(sts==TB_OK) ;
	    printf("%d %s\n",i+1,result) ;
	}
	break ;
      default:
	sts = TB_E_NOTYPE ;
    }

    sts = tb_free(table) ;
    assert(sts==TB_OK) ;

}
#endif


#ifdef TB2TBL
 
main(argc,argv) int argc ; char *argv[] ;
{
    table_t *table ;
    int sts ;
    int i ;

    for(i=1 ; i<argc ; i++) {
	sts = tb_read(argv[i],&table) ;    assert(sts==TB_OK) ;
	sts = tb_print(table,"",0) ;         assert(sts==TB_OK) ;
	sts = tb_free(table) ;             assert(sts==TB_OK) ;
    }
}
#endif
